//include ROS
#include "ros/ros.h"

//include the header file for the Twist message (used to specify velocity commands)
#include "geometry_msgs/Twist.h"

//include the header file for the turtlesim/Pose message (what the simulator publishes)
#include "turtlesim/Pose.h"

//TO DO: include the  header for our own .srv file
#include "first_package_cs378/TurtleCommand.h"

#define PI 3.14159265359

//global publisher for the turtle's velocity
ros::Publisher velocity_pub;

//global variable storing the turtle's current pose
turtlesim::Pose current_pose;
turtlesim::Pose last_pose;
bool have_last_pose = false;
bool heard_pose = false;

float distance_traveled = 0.0;


void pose_cb(const turtlesim::Pose::ConstPtr& msg){
	current_pose = *msg;
	
	//compute distance between last pose and current pose
	if (have_last_pose){
		
	}
	
	
	last_pose = current_pose;
	have_last_pose = true;
	heard_pose = true;
}

void waitForPose(){
	float framerate = 40.0;
	ros::Rate loop_rate(framerate);
	heard_pose = false;
	
	while (!heard_pose){
		ros::spinOnce();
		loop_rate.sleep();
	}
	
}

bool closeEnough(float angle1, float angle2, float eps){
	float d1 = fabs(angle1 - angle2);
	float d2 = fabs(angle1 - angle2);
	
	ROS_INFO("%f, %f",d1,d2);
	
	if (d1 < eps || d2 > (2*PI - eps)){
		return true;
	}
	else {
		return false;
	}
}

//assumes desired orientation is 0 to 2*PI (i.e., counter-clockwise)
void turn(float angular_vel, float desired_angle, float error_threshold){
	
	
	float framerate = 40.0;
	ros::Rate loop_rate(framerate);
	
	
	while (ros::ok()){

		//listen for incomming messages
		ros::spinOnce();
		
		//convert current angle to be positive
		float current_angle = current_pose.theta;
		if (current_angle < 0){
			current_angle = 2*PI + current_angle;
		}
		
		//get counter-clockwise and clockwise difference
		float cc_difference, cl_difference;
		if (desired_angle > current_angle){
			cc_difference = (desired_angle - current_angle);
			cl_difference = (2*PI-desired_angle+current_angle);
		}
		else {
			cc_difference = (2*PI-current_angle+desired_angle);
			cl_difference = (current_angle - desired_angle);
		}
		
		//the smaller of these difference is the angular error
		float angular_error;
		if (cc_difference < cl_difference)
			angular_error = cc_difference;
		else angular_error = cl_difference;
		ROS_INFO("angular error = %f",angular_error);

		geometry_msgs::Twist v;
		
		//decide which way to turn
		if (cc_difference < cl_difference)
			v.angular.z = angular_vel;
		else 
			v.angular.z = -1*angular_vel;
			
		if ((angular_error) < error_threshold){
			v.angular.z = 0;
			velocity_pub.publish(v);
			break;
		}
		else {
			velocity_pub.publish(v);
		}
		
		loop_rate.sleep();
	}
}

void moveForward(float v_x, float target_x, float target_y, float error_threshold){
	geometry_msgs::Twist v;
	v.linear.x = v_x;

	float framerate = 40.0;
	ros::Rate loop_rate(framerate);
	
	float last_error = sqrt( pow(target_x-current_pose.x,2) + pow(target_y-current_pose.y,2) );
	
	while (ros::ok) {
		
		ros::spinOnce();//sense own position
		
		//compute error
		float error = sqrt( pow(target_x-current_pose.x,2) + pow(target_y-current_pose.y,2) );
	
		ROS_INFO("error = %f",error);
		
		if (error < error_threshold){
			//stop
			v.linear.x = 0.0;
			velocity_pub.publish(v);
			break;
		}
		else {
			velocity_pub.publish(v);
		}
		
		//check if we're getting further from the target
		if (last_error < error){
			ROS_WARN("We're moving away from the target!!!");
			break;
		}
		last_error = error;
		
		loop_rate.sleep();
		ROS_INFO("Moving forward....");
	}
}


bool service_callback(
		first_package_cs378::TurtleCommand::Request  &req,
        first_package_cs378::TurtleCommand::Response &res)
{
	ROS_INFO("I got called!");
	ROS_INFO("The command I received is %s", req.command.c_str());
	
	if (req.command == "goto"){
		float x_t = req.x;
		float y_t = req.y;
		
		//update turtle's pose
		waitForPose();
		
		float dx = x_t - current_pose.x;
		float dy = y_t - current_pose.y;
		
		//get desired angle and convert it to be positive (counter-clockwise)
		float desired_angle = atan2(dy,dx);
		if (desired_angle < 0){
			desired_angle = 2*PI + desired_angle;
		}
		
		//get current angle and convert it to be positive (counter-clockwise)
		float current_angle = current_pose.theta;
		if (current_angle < 0){
			current_angle = 2*PI + current_angle;
		}
		
		ROS_INFO("dx = %f, dy = %f, d_angle = %f, c_angle =%f",dx,dy,desired_angle,current_angle);
		
		
		//now turn
		float angular_velocity = 1.0;
		turn(angular_velocity,desired_angle,0.025);
		
		//now go forward
		float linear_velocity = 1.0;
		moveForward(linear_velocity,x_t,y_t,0.25);
		
		//turn_and_move(x_t,y_t);
		
		res.result = true;
	}
  
  return true;
}

int main(int argc, char **argv){
	
	//declare our node to ROS
	ros::init(argc, argv, "turtle_controller_server");
	ros::NodeHandle n;
	
	//create the publisher for command velocities
	velocity_pub = n.advertise<geometry_msgs::Twist>("/turtle1/cmd_vel", 1);

	//subscribe to the turtle's position and orientation 
	ros::Subscriber sub_pose = n.subscribe("/turtle1/pose", 1, pose_cb );
   
	//TO DO: declare a service and write the callback function
    ros::ServiceServer service = n.advertiseService("/turtle_controller_server/command", service_callback);
   
    //finally, just spin (i.e., listen for service requests and messages)
	ros::spin();
	
	
	return 0;
}
